# -*- coding: utf-8 -*-

# In[1]:

import keras as K
import keras.layers as L
import tensorflow as tf
import scipy.io as sio
import argparse,os
import numpy as np
import h5py
import time
import sys
from data_util import * 
from sklearn.metrics import confusion_matrix  
from sklearn.metrics import classification_report 
from keras.preprocessing.image import ImageDataGenerator
from keras.callbacks import ModelCheckpoint,TensorBoard
from keras.utils import plot_model
# In[3]:

parser=argparse.ArgumentParser()
parser.add_argument('--NUM_EPOCH',
                    type=int,
                    default=500,
                    help='number of epoch')
parser.add_argument('--mode',
                    type=int,
                    default=0,
                    help='train or test mode')
parser.add_argument('--train_KEY',
                    type=str,
                    default='data',
                    help='train data')
parser.add_argument('--test_KEY',
                    type=str,
                    default='hsi',
                    help='test data')
parser.add_argument('--ksize1',
                    type=int,
                    default='11',
                    help='patch row')
parser.add_argument('--ksize2',
                    type=int,
                    default='11',
                    help='patch column')

args=parser.parse_args()

# NUM_CLASS=15
# HEIGHT=349
# WIDTH= 1905
# NUM_CHN= 144  #10 for PCA a 10 channels data



    

if not os.path.exists('log/'):
    os.makedirs('log/')

model_name=os.path.join(weights_path+args.train_KEY+'.h5')
# In[4]:

def GW_net(input_spat):
    filters=[32,64,128,256]
    input_spat=L.Input((args.ksize1,args.ksize2,NUM_CHN))
    # 9*9
    # define hsi_spatial convolution　
    conv0_spat=L.Conv2D(filters[2],(3,3),padding='valid')(input_spat)
    conv0_spat=L.BatchNormalization(axis=-1)(conv0_spat)
    conv0_spat=L.Activation('relu')(conv0_spat)
    # 7*7
    # conv0_spat=L.Conv2D(filters[1],(1,1),padding='valid',activation='relu')(conv0_spat)
    # 7*7
    conv1_spat = L.Conv2D(filters[2], (3, 3), padding='valid')(conv0_spat)
    # 5*5
    # conv1_spat = L.BatchNormalization(
    #     axis=-1, momentum=0.98, gamma_regularizer=K.regularizers.L1L2(l2=1e-3))(conv1_spat)
    conv1_spat=L.Activation('relu')(conv1_spat)
    # conv2_spat=L.Conv2D(filters[3],(3,3),padding='valid',activation='relu')(conv1_spat)
    # 3*3
    conv3_spat=L.Conv2D(filters[2],(1,1),padding='valid',activation='relu')(conv1_spat)

    # conv4_spat=L.Deconv2D(filters[1],(5,5),padding='valid',activation='relu')(conv3_spat)#!!!!!!!!!!!!!
    conv5_spat=L.Conv2D(filters[1],(5,5),padding='valid',activation='relu')(input_spat)#!!!!!!!!!!!!!
    conv6_spat=L.Conv2D(filters[1],(3,3),padding='valid',activation='relu')(conv0_spat)#!!!!!!!!!!!!!
    conv3_spat=L.concatenate([conv3_spat,conv6_spat,conv5_spat],axis=-1)#!!!!!!!!!!!!!
    conv3_spat=L.BatchNormalization(axis=-1)(conv3_spat)#!!!!!!!!!!!!!
    conv7_spat=L.Flatten()(conv3_spat)
    logits=L.Dense(NUM_CLASS,activation='softmax')(conv7_spat)
    model = K.models.Model([input_spat], logits)
    opti = K.optimizers.SGD(lr=0.001, momentum=0.99, decay=1e-3)
    kwargs=K.backend.moving_averages
    model.compile(optimizer=opti,loss='categorical_crossentropy',metrics=['acc'],kwargs=kwargs)
    return model

def SMALL_net(input_spat):
    filters=[32,64,128,256]
    conv0_spat=L.Conv2D(filters[2],(3,3),padding='valid')(input_spat)
    conv0_spat=L.BatchNormalization(axis=-1)(conv0_spat)
    conv1_spat=L.Activation('relu')(conv0_spat)
    conv2_spat=L.Conv2D(filters[2],(1,1),padding='valid',activation='relu')(conv1_spat)
    conv3_spat=L.Activation('relu')(conv2_spat)
    conv7_spat=L.Flatten()(conv3_spat)
    logits=L.Dense(NUM_CLASS,activation='softmax')(conv7_spat)
    model = K.models.Model([input_spat], logits)
    opti = K.optimizers.SGD(lr=0.001, momentum=0.99, decay=1e-3)
    kwargs=K.backend.moving_averages
    model.compile(optimizer=opti,loss='categorical_crossentropy',metrics=['acc'],kwargs=kwargs)
    return model

def SPA_net():
    input_spat=L.Input((args.ksize1,args.ksize2,NUM_CHN))
    if args.ksize1==3:
        net=SMALL_net(input_spat)
    else:
        net = GW_net(input_spat)
    return net

def train(model,model_name):
    model_ckt = ModelCheckpoint(filepath=model_name, verbose=1, save_best_only=True)
    tensorbd=TensorBoard(log_dir='./log',histogram_freq=0, write_graph=True, write_images=True)
    train_data=read_data(PATH,TRAIN_H,'train',args.train_KEY)
    # train_labels=K.utils.np_utils.to_categorical(train_data.labels,NUM_CLASS)
    train_labels=K.utils.np_utils.to_categorical(train_data.labels)
    print('train hsi data shape:{}'.format(train_data.hsi.shape))
    print('{} train sample'.format(train_data.hsi.shape[0]))
    class_weights={}
    N=np.sum(train_data.labels!=0)
    for c in xrange(NUM_CLASS):
        n=1.0*np.sum(train_data.labels==c)
        item={c:n}
        class_weights.update(item)
    print class_weights
    model.fit([train_data.hsi],train_labels, 
             batch_size=BATCH_SIZE,
            #  class_weight=class_weights,
             epochs=args.NUM_EPOCH,
             verbose=1,
             validation_split=0.1,
             shuffle=True,
             callbacks=[model_ckt,tensorbd])
    model.save(os.path.join(model_name+'_'))

def test(model_name,hsi_data):
    model = SPA_net()
    model.load_weights(model_name)
    pred=model.predict([hsi_data],batch_size=BATCH_SIZE)
    return pred
    

def main(mode=0,show=False):
    if args.mode==0:
        start_time=time.time()
        model=SPA_net()
        plot_model(model,to_file='model.png',show_shapes=True)
        train(model,model_name)
        duration=time.time()-start_time
        print duration
        # train_generator(model,model_name)
    else:
        # test_data,idx,idy=read_data(PATH,CTEST,'test',args.test_KEY)
        # prediction=test(model_name,test_data.hsi,test_data.labels)
        start_time=time.time()
        prediction=np.zeros(shape=(1,NUM_CLASS),dtype=np.float32)
        idxx=np.zeros(shape=(1,),dtype=np.int64)
        idyy=np.zeros(shape=(1,),dtype=np.int64)
        labels=np.zeros(shape=(1,),dtype=np.int64)
        for iclass in xrange(1,NUM_CLASS+1):
            CTEST=os.path.join('test_data_H'+str(iclass)+'.mat')
            test_data,idx,idy=read_data(TEST_PATH,CTEST,'test',args.test_KEY)
            tmp1=np.array(test(model_name,test_data.hsi),dtype=np.float32)
            # print(np.argmax(tmp1,1))
            prediction=np.concatenate((prediction,tmp1),axis=0)
            idxx=np.concatenate((idxx,idx),axis=0)
            idyy=np.concatenate((idyy,idy),axis=0)
            tmp_label=test_data.labels
            labels=np.concatenate((labels,tmp_label),axis=0)
        prediction=np.delete(prediction,0,axis=0)
        duration=time.time()-start_time
        print duration
        idxx=np.delete(idxx,0,axis=0)
        idyy=np.delete(idyy,0,axis=0)
        labels=np.delete(labels,0,axis=0)
        f = open(os.path.join('prediction_'+args.train_KEY+'.txt'), 'w')
        n = prediction.shape[0]

        for i in xrange(n):
            pre_label = np.argmax(prediction[i],0)
            f.write(str(pre_label)+'\n')
        f.close()

        print(prediction.shape,labels.shape)
        print('OA: {}%'.format(eval(prediction,labels)))

        # generate classification map
        pred_map=generate_map(prediction,idxx,idyy)

        # generate confusion_matrix
        prediction=np.asarray(prediction)
        pred=np.argmax(prediction,axis=1)
        pred=np.asarray(pred,dtype=np.int8)
        print confusion_matrix(labels,pred)

        # generate accuracy
        print classification_report(labels, pred)
if __name__ == '__main__':
    main()
    
